#pragma once

#include "forward.hpp"

namespace render::vk {

struct Subpass {
    VkShaderModule vsShader;
    VkShaderModule gsShader = NULL;
    VkShaderModule fsShader;
    VkPipelineLayout layout;
    VkPipeline pipeline = NULL;
};

class Pipeline final {
public:
    Pipeline(VkDevice, const PhysicalDeviceInfo&, const renderOptions&);
    ~Pipeline();


    // Universe renderpass
    constexpr VkRenderPass getRenderPass() const { return renderPass; }
    // Voxels (world & entity) passes descriptor set
    constexpr VkDescriptorSetLayout getVoxelDescriptorSet() const { return voxelDescriptorSet; }
    constexpr const Subpass& getTerrainPass() const { return terrainPass; }
    constexpr const Subpass &getEntitiesPass() const { return terrainPass; }
    constexpr VkDescriptorSetLayout getIndicDescriptorSet() const { return indicDescriptorSet; }
    constexpr const Subpass &getIndicPass() const { return indicPass; }
    constexpr const Subpass& getTransparentTerrainPass() const { return transparentTerrainPass; }
    constexpr VkDescriptorSetLayout getSkyDescriptorSet() const { return skyDescriptorSet; }
    constexpr const Subpass& getSkyPass() const { return skyPass; }

    constexpr VkRenderPass getUIRenderPass() const { return uiRenderPass; }

private:
    VkDevice device;

    VkRenderPass renderPass;

    VkDescriptorSetLayout voxelDescriptorSet;
    VkDescriptorSetLayout indicDescriptorSet;
    VkDescriptorSetLayout skyDescriptorSet;

    Subpass terrainPass;
    Subpass indicPass;
    Subpass transparentTerrainPass;
    Subpass skyPass;

    VkRenderPass uiRenderPass;
};
}