#include "Universe.hpp"

#include "../contouring/FlatDualMC.hpp"
#include "core/world/Elements.hpp"

using namespace world::client;

Universe::Universe(const std::string& ct): world::Universe(), contouring(std::make_unique<contouring::FlatDualMC>(ct)) { }
Universe::~Universe() { }

world::transform Universe::getAbsolute(const relative_transform& rtf) const {
    return getElements()->computeAbsolute(rtf.parent, rtf.relative);
}

world::Universe::query_result Universe::raycast(const geometry::Ray &ray) const {
    return Raycast(*getElements(), ray, true, getPlayerId());
}
/// Check for entities in range
bool Universe::isRangeFree(const node_voxel_ref &center, const action::Volume& volume) const {
    return IsRangeFree(*getElements(), center, geometry::Volume{action::ToGeometry(volume.shape), volume.radius});
}

world::relative_transform Universe::tryMovePlayer(const relative_transform& from, const offset_pos& move) const {
    const auto dir = glm::normalize(move);
    const auto l = getElements();
    const auto abs = l->computeAbsolute(from.parent, from.relative);
    //MAYBE: abs.rotation * dir
    const auto collide = Raycast(*l, geometry::Ray(abs.position + (world_pos)dir, dir, glm::length(move) + 1), true, getPlayerId());
    if (std::holds_alternative<voxel_query_target>(collide) || std::holds_alternative<area_query_out_of_range>(collide)) {
        return from;
    }
    return relative_transform{from.parent, transform(from.relative.position + (world_pos)move, from.relative.rotation)};
}