#pragma once

#include <fstream>
#include "../memory.hpp"
#include "./logger.hpp"

namespace io {

/// Map vector to istream
struct vec_istream: std::streambuf {
    vec_istream(const std::vector<char> &vec) {
        this->setg((char*)&vec[0], (char*)&vec[0], (char*)&vec[0] + vec.size());
    }
};

using namespace memory;
/// File loader
class file_content: public std::vector<char> {
public:
    /// Read first
    file_content(const std::vector<std::string>& paths, read_view prefix = read_view(nullptr, 0)):
        std::vector<char>(), prefix_size(prefix.size())
    {
        std::ifstream is = [&]() {
            for(auto& path: paths) {
                std::ifstream is(path, std::ios::in | std::ios::binary | std::ios::ate);
                if(is.good()) {
                    return is;
                }
                is.close();
            }
            FATAL("File not found " << paths.back());
        }();
        const auto end = is.tellg();
        is.seekg(0, std::ios::beg);
        resize(end - is.tellg() + prefix_size);
        memcpy(data(), prefix.ptr, prefix_size);
        is.read(data() + prefix_size, size());
        is.close();
    }

    size_t prefix_size;

    /// Data after prefix
    read_view content() const {
        return read_view((const uint8_t*)data() + prefix_size, size() - prefix_size);
    }
};

class ifstream: public std::ifstream {
public:
    ifstream(const std::string& path, std::ios_base::openmode mode = std::ios_base::in):
        std::ifstream(path, mode) { }

    template<typename D>
    void read_cast(D& d) {
        std::ifstream::read(reinterpret_cast<char *>(&d), sizeof(d));
    }
};

class ofstream: public std::ofstream {
public:
    ofstream(const std::string& path, std::ios_base::openmode mode = std::ios_base::out):
        std::ofstream(path, mode) { }

    template<typename D>
    void write_cast(const D& d) {
        std::ofstream::write(reinterpret_cast<const char *>(&d), sizeof(d));
    }
};

}