#pragma once

#include <variant>
#include "./Area.hpp"
#include "./Voxel.hpp"
#include "../geometry/Shapes.hpp"

/// Client action on world
namespace world::action {

namespace part {
struct Ping { };
}

struct Fill: part::Ping {
    Fill(const node_voxel_ref &pos, const Voxel &val): pos(pos), val(val) {}

    const node_voxel_ref pos;
    const Voxel val;
};

enum class Shape: uint8_t {
    Cube,
    RawSphere,
    SmoothSphere,
    /*CylinderX,
    CylinderY,
    CylinderZ,
    ConePX,
    ConeNX,
    ConePY,
    ConeNY,
    ConePZ,
    ConeNZ,
    */
};
static const std::array<std::string, 3> SHAPES = {"Cube", "Raw Sphere", "Smooth Sphere"};
static _FORCE_INLINE_ geometry::Shape ToGeometry(Shape shape) {
    switch (shape) {
    case Shape::RawSphere:
        return geometry::Shape::Sphere;
    case Shape::SmoothSphere:
        return geometry::Shape::Sphere;
    default:
        return geometry::Shape::Cube;
    }
}

struct Volume {
    Volume(Shape shape, uint8_t radius):
        shape(shape), radius(radius) { }

    const Shape shape;
    const uint8_t radius;
};

struct FillShape: Fill, Volume {
    FillShape(const node_voxel_ref &pos, const Voxel &val, Shape shape, uint8_t radius):
        Fill(pos, val), Volume(shape, radius) { }
};
struct Move: part::Ping {
    Move(const relative_pos& pos): pos(pos) { }

    const relative_pos pos;
};
struct Message: part::Ping {
    Message(const std::string& text): text(text) { }

    const std::string text;
};

using packet = std::variant<Move, Message, Fill, FillShape>;
}