#pragma once

#include "./actions.hpp"
#include "../geometry/Faces.hpp"
#include "../queue/safe_queue.hpp"
#include "../utils/mutex.hpp"

namespace world {

/// Link between world::server::SharedUniverse and world::client::LocalUniverse
struct server_handle {
    bool running = false;
    node_id entityId;
    mutex::shared_guard_ref<world::Elements> elements;

    std::string playerName;
    std::function<void(const world::action::packet &packet)> emit;
    std::optional<relative_transform> teleport;
    data::safe_queue<std::string> messages;
};

struct AbstractServerFactory {
    virtual ~AbstractServerFactory() { }
    virtual server_handle* run() = 0;
    virtual void onGui() = 0;
};
}